/*
 *  UnDo/ReDo
 */

#include "head.h"

unsigned MAXUNDO=100;
unsigned group_undo=1;

undo_el::undo_el()
{   mode=NONE; ptr=loc=NULL; x=0;
    Lines=Lineno=0;
    Next=Prev=NULL;
};

undo_el::undo_el (undo_el &el)
{   mode=el.mode;
    ptr=el.ptr;
    loc=el.loc;
    x=el.x;
    Lines=el.Lines;
    Lineno=el.Lineno;
    Next=Prev=NULL;
};

undo_el::undo_el(MODE m, int xx, int lno, int lines)
{   mode=m; ptr=loc=NULL; x=xx;
    Lineno=lno; Lines=lines;
    Next=Prev=NULL;
};

undo_el::undo_el(MODE m, VStr *p, VStr *l, int xx, int lno, int lines)
{   mode=m; ptr=p; loc=l; x=xx;
    Lineno=lno; Lines=lines;
    Next=Prev=NULL;
};

void undo_el::del()
{       if (ptr!=NULL) DisposeVStr(ptr);
        mode=NONE;
        loc=NULL;
} //undo_el::~undo_el

undo_b::undo_b():maxundo(MAXUNDO)
{       beg=fin=new undo_el;
        rbeg=rfin=new undo_el;
        cnt=0;
} //undo_b::undo_b

undo_b::undo_b(unsigned &max):maxundo(max)
{       beg=fin=new undo_el;
        rbeg=rfin=new undo_el;
        cnt=0;
} //undo_b::undo_b

undo_b::~undo_b()
{       while (beg)
        {       undo_el *p=beg->Next;
                beg->del();
                delete beg;
                beg=p;
        }
        while (rbeg)
        {       undo_el *p=rbeg->Next;
                rbeg->del();
                delete rbeg;
                rbeg=p;
        }
} //undo_b::~undo_b

//    undo
int undo_b::get (undo_el &el)
{       if (fin==beg) return 0;
        el= *fin;
        fin=fin->Prev;
        delete fin->Next;
        fin->Next=NULL;
        cnt--;
        return 1;
} //undo_b::get

//    undo
void undo_b::put (undo_el el)
{	if (group_undo && el.mode==REPL && fin->mode==REPL
		&& fin->Lineno==el.Lineno && fin->loc==el.ptr)
	{	fin->loc=el.loc;
                el.del();
		return;
	}
	fin->Next=new undo_el(el);
	fin->Next->Prev=fin;
	fin=fin->Next;
	cnt++;
	if (cnt<=maxundo) return;
	//   ( ) undo-
	undo_el *p=beg, *pp=NULL;
	for (unsigned i=0; p->Next&&(i<cnt-maxundo); i++)
        {       p=p->Next;
                switch (p->mode)
                { case ENDGROUP:
                  case ENDGROUP2:
                        pp=p;
                        break;
                  case GROUP:
                  case GROUP2:
                        pp=NULL;
                        break;
                  default:;
                }
        } //for
        if (pp) p=pp;   //  ,    
        while (beg->Next!=p)
        {       pp=beg->Next;
                beg->del();
                delete beg;
                beg=pp;
                cnt--;
        } //while
} //undo_b::put

int Edit::rest0 (undo_el &el)
{       if (!UndoBuf->get(el)) return 0;
        switch (el.mode)
        { case DEL:
                if (el.ptr==NULL) return 0;
                if (el.loc==NULL) return 0;
                el.ptr->Prev=el.loc->Prev;
                el.ptr->Next=el.loc;
                if (el.ptr->Prev!=NULL)
                        el.ptr->Prev->Next=el.ptr;
                if (el.ptr->Next!=NULL)
                        el.ptr->Next->Prev=el.ptr;
                if (First==el.loc) First=el.ptr;
                break;
          case REPL:
                if (el.ptr==NULL) return 0;
                if (el.loc==NULL) return 0;
                repline(el.loc,el.ptr,1);
                break;
          case INS:
                if (el.loc==NULL) return 0;
                repline(el.loc->Prev,NULL,1);
                break;
          default:;
        }
        return 1;
} //Edit::rest0

int Edit::restore()
{ undo_el el;
  int res;
  int i,m;
        res=rest0(el);
        if (!res) return 0;
        m=el.mode;
ree:    if (m==GROUP2)
           res=rest0(el);
        if (el.mode==GROUP)
           do   res=rest0(el);
           while (res&&(el.mode!=ENDGROUP));
        if (!res) return 0;
        Lines=el.Lines;
        Lineno=el.Lineno;
        if (!GapCnt && (Lineno<0)) Lineno+=Lines+1;
        if (Lineno<0)
        {   for (Cur=Last,i= -1; Cur!=First; Cur=Cur->Prev,i--)
                if (i==Lineno) break;
        }
        else
        {   for (Cur=First,i=1; Cur!=Last; Cur=Cur->Next,i++)
                if (i==Lineno) break;
        }
        modified=0;
        Mfin=xpos=el.x;
        ypos=calcy(Cur);
        i=calcy(el.ptr);
        corrxoffs();
        if (ypos<0)
        { Scr=Cur; ypos=0;
          DrawView();
        }
        else if (oldxoffs!=xoffset)
        { DrawView();
          oldxoffs=xoffset;
        }
        else if (MarkBlk&&vert)
          DrawView();
        else
        switch (el.mode)
        { case REPL:
            if (i<0) break;
            sout(0,i+A.Y,GetVStr(el.ptr),el.ptr==Last,
                belong(Cur,MarkBlk,el.ptr,vert));
            break;
          case INS:
          { i=calcy(el.loc);
            if (i<0) break;
            scrdn(i);
            VStr *p=getlowline();
            sout(0,B.Y,p?GetVStr(p):(char *)"",p==Last,belong(Cur,MarkBlk,p,vert));
            break;
          }
          case DEL:
            if (i<0) break;
            scrup(i);
            sout(0,i+A.Y,GetVStr(el.ptr),el.ptr==Last,
                belong(Cur,MarkBlk,el.ptr,vert));
            break;
          case ENDGROUP:
            if (m==GROUP)
                DrawView();
            break;
          default:;
        } //switch
        if ((m==GROUP2) && (el.mode!=ENDGROUP2)) goto ree;
        return res;
} //Edit::restore
